"""utils.py : Utilities for eXistDA Product"""
# -*- coding: utf-8 -*-

#    This file is part of eXistDA.
#
#    Copyright (C) 2004  Sebastien PILLOZ - Ecole Normale Superieure de Lyon
#
#    eXistDA is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


__doc__ = """Utilities for eXistDA Product"""

import types, urllib
from App.Dialogs import MessageDialog
import xml.dom.minidom
import htmlentitydefs
from zLOG import LOG, INFO

# Changeable constants
USE_CDOMLETTE       = True                               # If set to True, xslt.py will use 4Suite cDomlette instead of Domlette to make transformations.
USE_EXPAT           = True                               # Use expat to compute dictionnaries if available
USE_XSLTPROC        = True                               # Use XSLTPROC if available in place of 4Suite
XSLTPROC            = "xsltproc"                         # Executable for xsltproc, with full path if needed


# Global constants for eXistDA and syndicator
DEBUG               = 0                                  # Set to 1,2,3 to have debug messages on the console (or log) (3 is the most verbose), or 0 to disable
THIS_SCRIPT         = "EXISTDA.utils.py"                 # Name of this script
CODE_ENCODING       = 'UTF-8'                            # Format in which the code itself is encoded
OBJ_ONLY            = 'object_only'                      # dict key for eXistDA._getAnythingAsString object_only parameter
VERBOSE             = 'verbose'                          # dict key for query
XSLT_TIMEOUT        = 20                                 # Timeout in secs for xsltproc


# Dict keys for extra parameters
DICT_STYLESHEET_PARAM    = "stylesheet_param"
DICT_STYLESHEET_BASE_URI = "xsl_base_uri"

# Zope constants
EXISTDA_METATYPE    = "eXist Database Adapter"           # Metatype of the database adapter
EXISTDASYN_METATYPE = "eXist DB syndicator"              # Metatype of the syndicator
EXISTMETHOD_METATYPE= "eXist Method"                     # Metatype of an xquery method
EXISTDACACHEMANAGER_METATYPE = "eXistDA Cache Manager"   # Metatype of a cache manager
REXISTDA_METATYPE   = "eXist Remote Database Adapter"    # Metatype of the remote database adapter

# Cache manager types
RAMCACHE            = "RAM Cache Manager"                # Cache manager in RAM
ZODBCACHE           = "ZODB Cache Manager"               # Cache manager within the ZODB
FSCACHE             = "FileSystem Cache Manager"         # Cache manager on the file system
ALL_CACHE_TYPES = [RAMCACHE, ZODBCACHE, FSCACHE]         # Don' forget to add any cache manager you need in this list

# Defaults
DEFAULT_PP          = 1                                  # Default for pretty print (0 or 1)
EXIST_ROOT_COLL     = '/db'                              # Root collection of the DB (usually '/db') !!!! NO ENDING / !!!!
DEFAULT_ENCODING    = 'UTF-8'                            # Default encoding if nothing is passed during init or mod
DEFAULT_PORT        = '8080'                             # Default eXist XMLRPC server's port if nothing is passed during init or mod
DEFAULT_CACHEMANAGER_TYPE = RAMCACHE                     # Default cache manager type
DEFAULT_CACHE_MAXSIZE = 10                               # Default cache max size (in MB)
DEFAULT_XMLRPC_PATH = '/xmlrpc'                          # Default address of eXist embedded XMLRPC server
DEFAULT_ZOPE_XMLRPC_PATH = '/existda'                    # Default address of an eXistDA on a remote Zope server


# Headers for URIs
HEADER_HTTP_URI     = "http://"                          # Beginning of an HTTP URI (usually 'http://')
HEADER_EXIST_URI    = "xmldb:exist://"                   # Beginning of an eXist URI (usually 'xmldb:exist://')
HEADER_FILE_URI     = "file://"                          # Beginning of a file URI (usually 'file://')

# Parameters for XML options
OK_STYLESHEET       = "stylesheet"                       # key of the dict containing the XSL string which renders a doc
OK_INDENT_SPACES    = "indent-spaces"                    # key of the dict containing the space preservation value
OK_ENCODING         = "encoding"                         # key of the dict containing the encoding value
OK_INDENT           = "indent"                           # key of the dict containing the indentation presentation value
OK_STYLESHEET_PARAM = "stylesheet-param"                 # key of the dict containing the XSL parameters
OK_NAMESPACES       = "namespaces"                       # key of the dict containing the namespaces used by an XPath query

KEY_VARIABLES       = "variables"                        # key of the dict used to pass variables to xqueries

# Permissions
PERM_WRITE_ACCESS   = "Write to XMLDB"                   # permission used to have a write access to the DB
PERM_READ_ACCESS    = "View"                             # permission used to have a read access to the DB
PERM_MANAGE_ACCESS  = "View management screens"          # permission used to have a management rights to the DB

# Exceptions throwned by the objects
EXC_DOC_MUST_BE_ON_EXIST_SERVER = "The document must be on eXist, so you must give an eXist URI !"
EXC_CANT_SAVE                   = "The document can't be saved."
EXC_NOT_RESOURCE                = "The given path is not a resource of the DB."
EXC_NOT_STRING                  = "The object can't be returned as a string."
EXC_NOT_THIS_DB                 = "The URI passed is not an URI which represents this object !"
EXC_NO_4SUITE                   = "Can't find 4Suite in your python/lib. External XSLT won't be available"
EXC_MUST_HAVE_A_DA              = "The result must have an eXist DA."
EXC_NOT_DIR                     = "The given path is not a directory on the filesystem."
EXC_NOT_CACHE_TYPE              = "This cache type doesn't exist."



def hasExpat():
    """Returns True if Expat module is available"""
    try:
        import xml.parsers.expat
        return True
    except:
        return False

def hasXSLTProc():
    """Returns True is xsltproc is installed and in the path"""
    try:
        from os import popen
        if popen(XSLTPROC) != None:
            return True
        else:
            return False
    except:
        return False


def isHTTPURI(URI):
    """Verifies if 'URI' is an HTTP URI (ie beggining with 'HEADER_HTTP_URI').

* Returns 1 if 'URI' is an HTTP-like URI. 0 otherwise

* 'URI' : the string to test"""

    try:
        if type(URI) != types.StringType:
            return 0
    except:
        pass

    try:
        if URI[:len(HEADER_HTTP_URI)] == HEADER_HTTP_URI:
            return 1
    except:
        pass

    return 0


def iseXistURI(URI):
    """Verifies if 'URI' is an eXist URI (ie beggining with HEADER_EXIST_URI).

* Returns 1 if 'URI' is an eXist-like URI. 0 otherwise

* 'URI' : the string to test"""

    try:
        if type(URI) != types.StringType:
            return 0
    except:
        pass

    try:
        if URI[:len(HEADER_EXIST_URI)] == HEADER_EXIST_URI:
            return 1
    except:
        pass

    return 0

def isFileURI(URI):
    """Verifies if 'URI' is a file URI (ie beggining with HEADER_FILE_URI).

* returns 1 if 'URI' is a file-like URI. 0 otherwise

* 'URI' : the string to test"""

    try:
        if type(URI) != types.StringType:
            return 0
    except:
        pass

    try:
        if URI[:len(HEADER_FILE_URI)] == HEADER_FILE_URI:
            return 1
    except:
        pass

    return 0

def is4SuiteAvailable():
    """Returns 1 if 4Suite can be used to make external XSLT, 0 otherwise."""
    try:
        from Ft.Xml.Xslt import Processor
        from Ft.Xml.InputSource import DefaultFactory
        return 1 # True
    except:
        return 0 # False


def URIHasServerName(URI):
    """Returns True if the URI is containing a server name, False otherwise.

* 'URI' : a string containing the URI to scan."""

    if URI[:len(HEADER_EXIST_URI)+1] == HEADER_EXIST_URI + '/':
        return 0

    elif URI[:len(HEADER_HTTP_URI)+1] == HEADER_HTTP_URI + '/':
        return 0

    return 1


def getServerNameFromURI(URI):
    """Get the server name from an URI.

* Returns a string containing the name of the [eXist | HTTP] server corresponding to the URI (with the port number). Returns empty string if none."""

    if iseXistURI(URI) or isHTTPURI(URI):
        try:
            if URIHasServerName(URI.strip()):
                return URI.strip().split('//')[1].split('/',1)[0]
            else:
                return ""
        except:
            return ""

    return ""


def getRscNameFromURI(URI):
    """Get the resource name from an URI.

* Returns the name of the [eXist | HTTP] resource corresponding to the URI. Returns empty string if none."""

    if iseXistURI(URI) or isHTTPURI(URI):
        try:
            if URIHasServerName(URI.strip()):
                return "/" + URI.strip().split('//')[1].split('/',1)[1]
            else:
                return "/" + URI.strip().split('///')[1]
        except:
            return ""

    return ""


def convert(text, encoding_from, encoding_to):
    """Converts 'text' , encoded in 'encoding_from' to the encoding 'encoding_to'.

* returns the text converted into 'encoding_to' (if necessary) or raise an exception if something weird happened

* 'text' : a string containing the text to convert

* 'encoding_from' : a string containing the encoding of the text to convert

* 'encoding_to' : a string containing the encoding to convert the text to"""

    if encoding_from.lower().strip() != encoding_to.lower().strip():
        return unicode(text, encoding_from).encode(encoding_to)

    return text


def getInvalidMsg(server='', port='', username='', password='', encoding=''):
    """Verifies if the parameters passed to create or modify an eXistDA object are valid for use. Not all is verified...

* Returns 'None' if no error is found. Returns a MessageDialog object if something bad happened."""

    # We can't permit localhost or 127.0.0.1 for server address because of syndication
    # which needs to return the full qualified server name.
    if str(server).strip() == "ocalhost" or str(server).strip() == "127.0.0.":
        return MessageDialog(
            title = 'Error',
            message = "Server name can't be %s, it must be the full qualified name of the server. Changes not saved." % str(server).strip(),
            action = './manage_main')

    # Server name can't be empty
    if str(server).strip() == "":
        return MessageDialog(
            title = 'Error',
            message = "Server name can't be empty.",
            action = './manage_main')

    # Port must be an integer between 1 and 65535
    try:
        if int(str(port).strip()) < 1 or int(str(port).strip()) > 65535:
            # Not a port number...
            raise ValueError

    except ValueError:
        return MessageDialog(
            title = 'Error',
            message = "Port number must be an integer between 1 and 65535.",
            action = './manage_main')

    return None

# Utilities to get the DOM of an XML document into a python dictonary

def recurse_getDict(noeud, encoding, lst_nom_noeuds, cedico):
    """Returns a part of a dict which keys are XML tags and their values."""
    cle = '/'.join(lst_nom_noeuds).encode(encoding)

    for sousnoeud in noeud.childNodes:
        if sousnoeud.nodeType == sousnoeud.TEXT_NODE:
            valeur = sousnoeud.data.encode(encoding).strip().replace('\n','')
            if valeur != '':
                if cedico.has_key(cle):
                    cedico[cle].append(valeur)
                else:
                    cedico.update({cle:[valeur]})
        else:
            lst_nvl_cle = lst_nom_noeuds + [sousnoeud.tagName.encode(encoding)]
            cedico = recurse_getDict(sousnoeud, encoding, lst_nvl_cle, cedico)

    return cedico


def getDict(fichier_string, encoding=DEFAULT_ENCODING):
    """Returns the dict by calling the recursive method recurse_getDict."""

    for i in htmlentitydefs.entitydefs:
        if fichier_string.find("&%s;" % i) != -1 and i != 'amp':
            # We don't replace &amp; because & is not supported by expat.
            fichier_string = fichier_string.replace("&%s;" % i, htmlentitydefs.entitydefs[i])

    try:
        # Try to use expat if available and asked by user
        if USE_EXPAT and hasExpat():
            import xml.parsers.expat
            d = XMLToDict(encoding)
            return d.parseXML(fichier_string)
        else:
            doc = xml.dom.minidom.parseString(fichier_string)
            return recurse_getDict(doc.documentElement, encoding, [doc.documentElement.tagName.encode(encoding)], {})
    except:
        doc = xml.dom.minidom.parseString(fichier_string)
        return recurse_getDict(doc.documentElement, encoding, [doc.documentElement.tagName.encode(encoding)], {})


def mergeDict(dict1, dict2):
    """Returns a new dictionary by merging 'dict1' and 'dict2'.

**WARNING** : the two dicts must have the same structure ! If 'dict1["my_key"] and 'dict2["my_key"]' exist and don't have the same type of value,
an error will be raised."""

    dict = dict1.copy

    for one_key in dict2.keys():
        if dict.has_key(one_key):
            dict[one_key] = dict[one_key] + dict2[one_key]
        else:
            dict[one_key] = dict2[one_key]

    return dict

class XMLToDict:
    """Class implementing expat parser to map an XML string to a python dictionary"""

    def _start_element(self, name, attrs):
        """What to do when processing an opening tag"""
        self.tags_stack.append(name)
        for k in attrs.keys():
            key = '/'.join(self.tags_stack).encode(self.encoding) + '/@' + k.encode(self.encoding)
            if self.xml_dict.has_key(key):
                if self.wrote_data[key]:
                    self.xml_dict[key][-1] += attrs[k].encode(self.encoding)
                else:
                    self.xml_dict[key].append(attrs[k].encode(self.encoding))
            else:
                self.xml_dict[key] = [attrs[k].encode(self.encoding)]
            self.wrote_data[key] = True

    def _end_element(self, name):
        """What to do when processing a closing tag"""
        key = '/'.join(self.tags_stack).encode(self.encoding)
        self.tags_stack.pop()
        self.wrote_data[key] = False

    def _char_data(self, data):
        """What to do when processing character data in a tag"""
        key = '/'.join(self.tags_stack).encode(self.encoding)
        if data.strip().replace('\n', '') != '':
            if self.xml_dict.has_key(key):
                if self.wrote_data[key]:
                    self.xml_dict[key][-1] += data.encode(self.encoding)
                else:
                    self.xml_dict[key].append(data.encode(self.encoding))
            else:
                self.xml_dict[key] = [data.encode(self.encoding)]
            self.wrote_data[key] = True


    def __init__(self, encoding=DEFAULT_ENCODING):
        """Construction of the class, with encoding awareness"""
        self.encoding = str(encoding)
        self.xml_dict = {}
        self.tags_stack = []
        self.attrs_dict = {}
        self.wrote_data = {}


    def parseXML(self, xml_string):
        """Parse 'xml_string' and returns a dictionary mapped on the XML DOM"""
        p = xml.parsers.expat.ParserCreate(self.encoding)

        p.StartElementHandler = self._start_element
        p.EndElementHandler   = self._end_element
        p.CharacterDataHandler= self._char_data
        p.buffer_text = True

        if (len(xml_string) > 0):
            p.Parse(xml_string)

        return self.xml_dict


# Message LOG at startup
if hasExpat():
    if USE_EXPAT:
        LOG(THIS_SCRIPT, INFO, "Expat is available and will be used. If you want to use minidom instead, please change USE_EXPAT to False in %s." % THIS_SCRIPT)
    else:
        LOG(THIS_SCRIPT, INFO, "Expat is available but won't be used. If you want to use expat, please change USE_EXPAT to True in %s." % THIS_SCRIPT)
else:
    LOG(THIS_SCRIPT, INFO, "Expat is **NOT** available. minidom will be used instead. Please install pyexpat (see http://pyxml.sourceforge.net) if you want to use it.")

if hasXSLTProc():
    if USE_XSLTPROC:
        LOG(THIS_SCRIPT, INFO, "xsltproc is available, and will be used. If you don't want to use xsltproc but 4Suite, please change USE_XSLTPROC to False in %s." % THIS_SCRIPT)
    else:
        LOG(THIS_SCRIPT, INFO, "xsltproc is available, but won't be used. If you want to use xsltproc and not 4Suite, please change USE_XSLTPROC to True in %s." % THIS_SCRIPT)
else:
    LOG(THIS_SCRIPT, INFO, "xsltproc is **NOT** available (xsltproc not installed or not in path). 4Suite will be used if found. Else, external XSLT won't be available for eXistDA.")

# End message LOG



"""eXistDAresult.py : a result for eXistDA queries, getDoc or whatever implies XML"""

# -*- coding: ISO-8859-1 -*-

#    This file is part of eXistDA.
#
#    Copyright (C) 2004  Sebastien PILLOZ - Ecole Normale Superieure de Lyon
#
#    eXistDA is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


__doc__ = """Results for eXistDA Product"""

import utils
import eXistDA

from types import StringType
from zLOG import LOG, INFO, DEBUG

# Constants
SCRIPT= 'EXISTDARESULT'
cDA   = 'DA'
cXML  = 'xml'
cHITS = 'hits'
cRES  = 'results'
cENC  = 'encoding'
cDOCB = 'docbase'
resACCEPT = [cDA, cXML, cHITS, cENC, cDOCB]
resSetACCEPT = [cRES, cENC]


class eXistDAresult:
    """A result object for eXistDA product. It contains XML and other infos"""

    def __init__(self, DA=None, xml='', docbase='', hits=-1, encoding=utils.DEFAULT_ENCODING):
        """Initialization of the class"""
        if DA != None:
            self.DA = DA
        else:
            raise utils.EXC_MUST_HAVE_A_DA

        self.xml = str(xml)
        self.hits = int(hits)
        self.docbase = "" #docbase
        self.encoding = encoding


    def __setattr__(self, name, value):
        """Attributes setting"""

        if name in resACCEPT:
            if name == cHITS:
                self.__dict__[name] = int(value)
            elif name == cDOCB:
                self.__dict__[name] = '' #value
            elif name == cENC:
                if self.__dict__.has_key(cXML) and self.__dict__.has_key(cENC):
                    self.xml = utils.convert(self.xml, self.encoding, value)
                self.__dict__[name]=value
            else:
                self.__dict__[name] = str(value)


    def __getattr__(self, name):
        """Attributes recovering"""
        if self.__dict__.has_key(name):
            if name == cXML:
                return str(self.__dict__[name])

            return self.__dict__[name]
        else:
            return None


    def __eq__(self, other):
        """Equality comparison"""
        if unicode(self.xml, self.encoding) == unicode(other.xml, other.encoding):
            return True

        return False


    def __neq__(self, other):
        """Not equality"""
        if (unicode(self.xml, self.encoding) == unicode(other.xml, other.encoding)) :
            return False

        return True


    def __add__(self, other):
        """Adds two results and returns a resultSet"""
        if isinstance(other, eXistDAresult):
            return eXistDAresultsSet(self, other)
        else:
            raise TypeError, 'Right operand must be a eXistDAresult'


    def __cmp__(self, other):
        """Comparison. By default, we test hits"""
        return cmp(self.hits, other.hits)


    def __str__(self):
        """String representation"""
        return str(self.xml)
    __repr__ = __str__

    def getDict(self):
        """Returns the dictionary representing the XML result"""

        return utils.getDict(self.xml, encoding=self.encoding)



class eXistDAresultsSet:
    """A results set"""

    def __init__(self, encoding=utils.DEFAULT_ENCODING):
        """Creates an instance from the class, with n DAresults passed in the creation"""
        self.encoding = encoding
        self.__dict__[cRES] = []


    def __getattr__(self, name):
        """Attributes recovering"""
        if self.__dict__.has_key(name):
            return self.__dict__[name]
        else:
            return None


    def __getitem__(self, index):
        """ """
        return self.results.__getitem__(index)


    def __setattr__(self, name, value):
        """Attributes setting"""
        if name in resSetACCEPT:
            if name == cRES:
                if self.__dict__.has_key(cENC):
                    self.__dict__[name] = [unicode(str(value)).encode(self.encoding)]
                else:
                    self.__dict__[name] = str(value)
            elif name == cENC:
                new_results = []
                if self.__dict__.has_key(cRES):
                    for result in self.__dict__[cRES]:
                        result.encoding = value # Converts result to a new encoding 'value'
                        new_results.append(result)
                    self.__dict__[cRES] = new_results
                    if utils.DEBUG > 2 :
                        LOG(SCRIPT, DEBUG, "New results for eXistDAResultsSet is : %s" % str(self.results))

                self.__dict__[name] = value


    def __str__(self):
        """String representation"""
        retour =""
        for r in self.results:
            retour = retour + str(r) + '\n'

        return retour
    __repr__ = __str__

    def __add__(self, other):
        """Adds two resultsSets. The encoding is the left operand's one"""
        if isinstance(other, eXistDAresultsSet):
            for result in other.results:
                result.encoding = self.encoding
                self.results += result
            return self
        else:
            raise TypeError, 'Right operand must be a eXistDAresultsSet'


    def __nonzero__(self):
        """Test the 0 value of the class"""
        if len(self.results) == 0:
            return False
        return True


    def __len__(self):
        """Number of items in the list"""
        return len(self.results)


    def count(self, x) :
        """Return the number of occurence of x in the resultsSet. x can be an eXistDAresult, an eXistDA object or an XML string"""
        if isinstance(x, eXistDA.eXistDA):
            listeDA = []
            for result in self.results:
                listeDA.append(result.DA)
            return listeDA.count(x)
        elif isinstance(x, eXistDAresult):
            return self.results.count(x)
        elif type(x) == StringType:
            listeXML = []
            for result in self.results:
                listeXML.append(result.xml)
            return listeXML.count(x)
        else:
            raise TypeError, 'Seeked object must be an eXistDAresult instance, an eXistDA instance, or an XML string'


    def append(self, other):
        """Appends a result to this resultsSet"""

        if isinstance(other, eXistDAresult):
            self.results.append(other)
        else:
            raise TypeError, 'Appended object must be a eXistDAresult'


    def remove(self, x):
        """Removes the first occurence of the result containing x in the list. x can be an eXistDAresult, an eXistDA object, or an XML string"""
        if isinstance(x, eXistDA.eXistDA):
            for result in self.results:
                if x == result.DA:
                    self.results.remove(result)
                    break
        elif isinstance(x, eXistDAresult):
            self.results.remove(x)
        elif type(x) == StringType:
            for result in self.results:
                if x == result.xml:
                    self.results.remove(result)
                    break
        else:
            raise TypeError, 'Removed object must be an eXistDAresult instance, an eXistDA instance, or an XML string'


    def index(self, x):
        """Returns the index of the first occurence of x in the _results list.  x can be an eXistDAresult, an eXistDA object, or an XML string"""
        if isinstance(x, eXistDA.eXistDA):
            for result in self.results:
                if x == result.DA:
                    return self.results.index(result)
        elif isinstance(x, eXistDAresult):
            return self.results.index(x)
        elif type(x) == StringType:
            for result in self.results:
                if x == result.xml:
                    return self.results.index(result)
        else:
            raise TypeError, 'Seeked object must be an eXistDAresult instance, an eXistDA instance, or an XML string'


    def sort(self, fn=None):
        """Sorts the _results by score. Use reverse to reverse the order"""
        if not (fn is None):
            self.results.sort(fn)
        else:
            self.results.sort()


    def reverse(self):
        """Reverse the order of the results list"""
        self.results.reverse()


    def getMergedXML(self, xmled=True):
        """Returns the result as a big simple XML file"""
        str_res = ''
        if xmled:
            str_res = '<?xml version="1.0" encoding="%s"?><results>' % self.encoding

        for res in self.results:
            str_res += "<result>%s</result>" % res.xml

        if xmled:
            str_res += "</results>"

        return str_res


    def listResultsDocs(self):
        """Returns a list containing the URIs of each doc verifying the query

**BROKEN** : always returns the name of the server and its port, but not the path to the file
        """

        lst_res = []
        for res in self.results:
            str_res = utils.HEADER_EXIST_URI + res.DA.server + ':' + res.DA.port + res.docbase
            lst_res.append(str_res)

        return lst_res


    def getDict(self):
        """Returns a list of dictionaries representing the xml resultsSet"""

        res_list = []
        for elem in self.results:
            res_list.append(elem.getDict())

        return res_list


    def getMergedDict(self, xmled=True):
        """Returns a dictionary of the merged version of the resultsSet"""

        return utils.getDict(self.getMergedXML(xmled), self.encoding)




# -*- extra stuff goes here -*-
"""__init.py__ : eXistDA and eXistDAsyn initialization module."""

# -*- coding: utf-8 -*-

#    This file is part of eXistDA.
#
#    Copyright (C) 2004  Sebastien PILLOZ - Ecole Normale Superieure de Lyon
#
#    eXistDA is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

__doc__ = """eXistDA, ReXistDA and eXistDAsyn initialization module."""

import eXistDA, eXistDAsyn, ReXistDA

def initialize(context):
    """Initializer called when used as a Zope 2 product.

    Initialization of the two classes used to instanciate the two kind of objects seen by zope : eXistDA and eXistDAsyn."""
    try:
        # register the DA
        context.registerClass (
          eXistDA.eXistDA,
          constructors = (
            eXistDA.manage_addeXistDAForm,
            eXistDA.manage_addeXistDA),
          icon='images/eXistDA.png'
        )
        # register the syndicator
        context.registerClass (
          eXistDAsyn.eXistDAsyn,
          constructors = (
            eXistDAsyn.manage_addeXistDAsynForm,
            eXistDAsyn.manage_addeXistDAsyn),
          icon='images/eXistDAsyn.png'
        )
        # register the remote DA
        context.registerClass (
          ReXistDA.ReXistDA,
          constructors = (
            ReXistDA.manage_addReXistDAForm,
            ReXistDA.manage_addReXistDA),
          icon='images/ReXistDA.png'
        )

        context.registerHelp()

        """# register the method
        context.registerClass (
          eXistMethod.eXistMethod,
          constructors = (
            eXistMethod.manage_addeXistMethodForm,
            eXistMethod.manage_addeXistMethod),
          icon='images/eXistMethod.png'
        )"""



    except:
        import sys, traceback, string
        type, val, tb = sys.exc_info()
        sys.stderr.write(string.join(traceback.format_exception(type, val, tb), ''))
        del type, val, tb


import eXistDAresult
from AccessControl import allow_class
allow_class(eXistDAresult.eXistDAresult)
allow_class(eXistDAresult.eXistDAresultsSet)



"""eXistDAsyn.py : eXistDA Syndicator product module."""

# -*- coding: ISO-8859-1 -*-

#    This file is part of eXistDA.
#
#    Copyright (C) 2004  Sebastien PILLOZ - Ecole Normale Superieure de Lyon
#
#    eXistDA is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

__doc__ = """eXistDAsyn product module."""

from App.special_dtml import HTMLFile, DTMLFile
from App.Dialogs import MessageDialog
from Persistence import Persistent
from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass
from zLOG import LOG, DEBUG, INFO

import OFS.SimpleItem, OFS.ObjectManager
import Acquisition, App
import AccessControl.Role
import xmlrpclib, types
import utils
from utils import iseXistURI, isHTTPURI
import eXistDA
import eXistDAresult

## Zope methods for management
manage_addeXistDAsynForm = HTMLFile('dtml/eXistDAsynAdd', globals())
def manage_addeXistDAsyn(self, id, title='', encoding=utils.DEFAULT_ENCODING, REQUEST=None):
    """Add an eXist syndicator object."""

    # Avoid empty encoding
    if encoding.strip() == '':
        encoding = utils.DEFAULT_ENCODING

    self._setObject(id, eXistDAsyn(id, title, encoding))
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)


class eXistDAsyn (OFS.ObjectManager.ObjectManager, Persistent, Acquisition.Implicit, AccessControl.Role.RoleManager, App.Management.Navigation):
    """An eXist syndicator object, containing many eXistDA objects."""

    meta_type = utils.EXISTDASYN_METATYPE

    isAnObjectManager=1

    manage_options =  OFS.ObjectManager.ObjectManager.manage_options + (
      {'label':'Properties', 'action':'manage_properties'},
      {'label':'Security', 'action':'manage_access'},
    )


    security = ClassSecurityInfo()
    security.declareObjectProtected("View")

    security.declareProtected('View management screens', 'manage_tabs', 'manage_main', 'manage_edit')
    security.declareProtected('Change permissions', 'manage_access')

    # Methods to create an eXistDA object in the syndicator
    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'addeXistDA')
    addeXistDA = eXistDA.manage_addeXistDAForm

    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'manage_addeXistDA')
    manage_addeXistDA = eXistDA.manage_addeXistDA
    ###########

    def __init__(self, id, title, encoding=utils.DEFAULT_ENCODING):
        """Initialization of the eXistDAsyn object."""
        self.id = str(id)
        self.title = str(title)
        self.encoding = str(encoding)


    index_html = HTMLFile('dtml/index_syn', globals())
    manage_main = OFS.ObjectManager.ObjectManager.manage_main
    manage_workspace = OFS.ObjectManager.ObjectManager.manage_workspace
    manage_properties = HTMLFile('dtml/eXistDAsynEdit', globals())

    meta_types = [{ 'name' : utils.EXISTDA_METATYPE, 'action' : 'addeXistDA', 'permission' : 'Add eXistDA' }]
    all_meta_types = meta_types

    ###################
    ### Internal methods

    security.declarePrivate('_findServerByURI')
    def _findServerByURI(self, URI):
        """Find the corresponding DA in the syndicator by its URI.

* if a corresponding DA is found, returns the object. Else returns 'None'

* 'URI' : the string containing the URI to test"""

        if utils.iseXistURI(URI):
            # we get the servername and the port number
            try:
                svrname,port = utils.getServerNameFromURI(URI).split(':')
                for DA in self.objectValues([utils.EXISTDA_METATYPE]):
                    if DA.server == svrname and DA.port == port and self.restrictedTraverse(DA.id) != None :
                        return DA
            except:
                svrname= utils.getServerNameFromURI(URI)
                for DA in self.objectValues([utils.EXISTDA_METATYPE]):
                    if DA.server == svrname :
                        return DA

        return None

    security.declarePrivate('_findServerById')
    def _findServerById(self, id):
        """Find the corresponding DA in the syndicator by its id.

* if a corresponding DA is found, returns the object. Else returns 'None'

* 'id' : the string containing the id to test"""

        for DA in self.objectValues([utils.EXISTDA_METATYPE]):
            if DA.id == id:
                return DA

        return None


    security.declarePrivate('_getIteratedServers')
    def _getIteratedServers(self, my_items):
        """Returns a list containing all the servers we want to iterate over (may be empty).

* my_items : None, string or list which represents all the server ids we want to iterate over"""

        try:
            if (my_items is not None):
                if type(my_items) == types.StringType and my_items.strip() != '':
                    server = self._findServerById(my_items.strip())
                    if server is None:
                        return []
                    else:
                        return [server]

                if type(my_items) == types.ListType and len(my_items) > 0:
                    my_list = []
                    for i in my_items:
                        server = self._findServerById(my_items.strip())
                        if server is not None:
                            my_list.append(server)

                        return my_list
        except:
            pass

        liste = []
        for DA in self.objectValues([utils.EXISTDA_METATYPE]):
            try:
                if self.restrictedTraverse(DA.id) != None:
                    liste.append(DA)
            except:
                pass

        return liste





    ###############################################################################
    ### XMLRPC & eXist methods


    security.declareProtected(utils.PERM_READ_ACCESS, 'getId')
    def getId(self):
        """Returns the id of the object"""
        return self.id

    security.declareProtected(utils.PERM_READ_ACCESS, 'getTitle')
    def getTitle(self):
        """Returns the title of the object"""
        return self.title

    security.declareProtected(utils.PERM_READ_ACCESS, 'getEncoding')
    def getEncoding(self):
        """Returns the encoding of the object"""
        return self.encoding


    security.declareProtected(utils.PERM_READ_ACCESS, 'listDAids')
    def listDAids(self):
        """Returns a list of the ids of the DA contained in the syndicator."""
        liste = []
        for DA in self._getIteratedServers(None):
            liste.append(DA.id)

        return liste

    security.declareProtected(utils.PERM_READ_ACCESS, 'listDAs')
    def listDAs(self):
        """Returns a list of DAs contained in the syndicator."""
        liste = []
        for DA in self._getIteratedServers(None):
            liste.append(DA)

        return liste


    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'isOnline')
    def isOnline(self,servers=None):
        """Gets running servers.

* Returns a list containing tuples (id of servers, 1 if online | 0 otherwise)."""

        liste_online = []

        for DA in self.objectValues([utils.EXISTDA_METATYPE]):
            liste_online.append(DA.isOnline())

        return liste_online


    security.declareProtected(utils.PERM_READ_ACCESS, 'getFilesFromQuery')
    def getFilesFromQuery(self, query, **kws):
        """Query all the DBs in the syndicator.

* Returns a list of tuples like this (eXistDA object, filename) where filename matches the query

* 'query' : a string containing an XPATH query or an XQuery

* '**kws' : see eXistDA for doc. Ignored for the moment, we'll see later how to do this."""

        liste = []
        ns = {}
        if kws.has_key(utils.OK_NAMESPACES):
            ns = kws[utils.OK_NAMESPACES]

        for DA in self.objectValues([utils.EXISTDA_METATYPE]):
            try:
                if kws.has_key(utils.KEY_VARIABLES):
                    for file in DA.getFilesFromQuery(query, object_only=1, variables=kws[utils.KEY_VARIABLES], namespaces=ns):
                        liste.append((DA, file))
                else:
                    for file in DA.getFilesFromQuery(query, object_only=1, namespaces=ns):
                        liste.append((DA, file))
            except:
                pass

        liste_def = []
        for res_A in liste[:-1]:
            suppr = False
            for res_B in liste[liste.index(res_A)+1:]:
                if res_A[0].server == res_B[0].server:
                    if len(res_A[1]) < len(res_B[1]):
                        suppr = True
            if suppr != True:
                liste_def.append(res_A)

        return liste_def


    security.declareProtected(utils.PERM_READ_ACCESS, 'query')
    def query(self, query, xmled=1, **kws):
        """Query all the DBs in the syndicator.

* Returns a 'eXistDAresultsSet' object containing the results for all DAs returned by the queries.

* 'query' : a string containing an XPATH query or an XQuery

* 'xmled' : 1 to return a valid xml string (with <?xml... declaration), 0 to return the result as it comes from eXist

* 'kws' : other parameters

  * 'verbose' : if set to 1, returns a list of lists '[[DA1,[res1, nb_hist_1, query_time_1]], [DA2,[res2, nb_hits_2, query_time_2], ...]]'

  * 'variables' : dict containing variables to pass to the xquery
  """

        if utils.DEBUG >= 3:
            LOG("eXistDAsyn.py - query :", DEBUG, "Entering query method")

        resSet = eXistDAresult.eXistDAresultsSet(encoding = self.encoding)

        ns = {}
        if kws.has_key(utils.OK_NAMESPACES):
            ns = kws[utils.OK_NAMESPACES]

        old_DAs = []
        for DA in self.objectValues([utils.EXISTDA_METATYPE]):
            if utils.DEBUG >= 2:
                LOG("eXistDAsyn.py - query :", DEBUG, "Querying %s" % DA.server)
            #try:
            DA_prot = self.restrictedTraverse(DA.id)
            if utils.DEBUG >= 2:
                LOG("eXistDAsyn.py - query :", DEBUG, "DA_prot for this DA is %s" % str(DA_prot))
            if not(DA_prot.server in old_DAs) and DA_prot != None:
                if kws.has_key(utils.KEY_VARIABLES):
                    thisDA_resSet = DA_prot.query(query, variables=kws[utils.KEY_VARIABLES], namespaces=ns)
                else:
                    thisDA_resSet = DA_prot.query(query, namespaces=ns)


                LOG("eXistDAsyn.py - query :", DEBUG, "resSet is %s" % str(resSet))
                LOG("eXistDAsyn.py - query :", DEBUG, "thisDA_resSet is %s" % str(thisDA_resSet))
                if thisDA_resSet is not None:
                    if resSet is not None:
                        resSet = resSet + thisDA_resSet
                    else:
                        resSet = thisDA_resSet


                if utils.DEBUG >= 3:
                    LOG("eXistDAsyn.py - query :", DEBUG, "Adding query result from %s to syndicator resultsSet")
                    LOG("eXistDAsyn.py - query :", DEBUG, "My new resultsSet is : %s " % resSet.getMergedXML())

                old_DAs.append(DA_prot.server)
            #except:
                #pass

        return resSet



    security.declareProtected(utils.PERM_READ_ACCESS, 'getDoc')
    def getDoc(self, doc_name, xsl='', pretty_print=utils.DEFAULT_PP, **kws):
        """Retrieve a document in one of our DBs.

* Returns a string containing the (XSL transformed if suited) XML document.

* 'doc_name' : a string containing the server, the full path and name of an existing document in a DB

* 'xsl' : a string containing the full path and name of an existing XSL document in the DB (same as doc for the moment). If an empty string is passed, no XSLT is done

* 'if pretty_print' == 1, then the returned string is "pretty printed", ie indented. Else, the string is returned as it is saved in the DB"""

        DA = self._findServerByURI(doc_name)
        if DA is not None:

            if kws.has_key(utils.DICT_STYLESHEET_PARAM):
                return utils.convert(DA.getDoc(doc_name, xsl, pretty_print, stylesheet_param=kws[utils.DICT_STYLESHEET_PARAM]), DA.encoding, self.encoding)

            else:
                return utils.convert(DA.getDoc(doc_name, xsl, pretty_print), DA.encoding, self.encoding)



        return ""


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'xupdate' )
    def xupdate(self, res_name, xupdate_doc):
        """Issue a xupdate command to update a document or a whole collection in the database. If 'res_name' is an eXist URI, only this database (if it's in the syndicator) will be "XUpdated". If no server name is given, the command is issued on all DA in the syndicator.

* Returns the number of server where documents have been modified or raise an exception if pb

* 'res_name' : the name (with complete path) of a collection or a document

* 'xupdate_doc' : a string containing a xupdate file
       """
        nb_servers = 0

        if iseXistURI(res_name) :
            DA = self._findServerByURI(res_name)
            if DA is not None:
                DA.xupdate(utils.getRscNameFromURI(res_name), utils.convert(xupdate_doc, self.encoding, DA.encoding))
                nb_servers = nb_servers + 1
        else:
            for DA in self.objectValues([utils.EXISTDA_METATYPE]):
                try:
                    DA.xupdate(res_name, utils.convert(xupdate_doc, self.encoding, DA.encoding))
                    nb_servers = nb_servers + 1
                except:
                    pass

        return nb_servers

    security.declareProtected( utils.PERM_READ_ACCESS, 'listColl' )
    def listColl(self, base_coll=utils.EXIST_ROOT_COLL, recursive=0, full_path=1):
        """Lists collections in one DA (if server name is passed) or all DA applicable.

'base_coll' : URI of an eXist server (full or not)

'recursive' : integer, 1 will look into all subcollections, 0 only in the collection passed

'full_path' : integer, 1 will print full_path, 0 not
"""

        liste = []
        for DA in self.objectValues([utils.EXISTDA_METATYPE]):
            try:
                liste.append((DA.id,DA.listColl(base_coll, recursive, full_path)))
            except:
                pass

        return liste


    ## End XMLRPC & eXist methods
    ##########################################################################################

    security.declareProtected( utils.PERM_MANAGE_ACCESS, 'manage_edit' )
    def manage_edit(self, title="", encoding=utils.DEFAULT_ENCODING, REQUEST=None):
        """Change values of parameters"""
        # Avoid empty encoding
        if encoding.strip() == '':
            encoding = utils.DEFAULT_ENCODING

        self.title = str(title)
        self.encoding = str(encoding)

        if REQUEST is not None:
            return MessageDialog(
              title = 'Edited',
              message = "Properties for %s changed." % self.id,
              action = './manage_main',
            )

InitializeClass(eXistDAsyn)



"""ReXistDA.py : Remote eXist Database Adapter product module for Zope."""

# -*- coding: ISO-8859-1 -*-

#    This file is part of eXistDA.
#
#    Copyright (C) 2004  Sebastien PILLOZ - Ecole Normale Superieure de Lyon
#
#    eXistDA is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


__doc__ = """ReXistDA product module."""

from App.special_dtml import HTMLFile
from App.Dialogs import MessageDialog
from Persistence import Persistent
from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass

import OFS.SimpleItem
import Acquisition
import AccessControl.Role
from zLOG import LOG, DEBUG

import xmlrpclib, urllib
from xmlrpclib import Binary
from types import StringType

from utils import iseXistURI, isHTTPURI
import eXistDA
import utils
import eXistDAresult
import xslt


## Zope method for management
manage_addReXistDAForm = HTMLFile('dtml/ReXistDAAdd', globals())
def manage_addReXistDA(self, id, title='', server='', port='', username='', password='', encoding=utils.DEFAULT_ENCODING, xmlrpc_path=utils.DEFAULT_ZOPE_XMLRPC_PATH, REQUEST=None):
    """Add an Remote eXist connector object."""
    # Avoid empty encoding
    if encoding.strip() == '':
        encoding = utils.DEFAULT_ENCODING

    msg = utils.getInvalidMsg(server, port, username, password, encoding)
    if msg is None :
        self._setObject(id, ReXistDA(id, title, server, port, username, password, encoding, xmlrpc_path))
        if REQUEST is not None:
            return self.manage_main(self, REQUEST)
    else:
        if REQUEST is not None:
            return msg

# The class which represents an remote eXist "proxy" or database adapter.
class ReXistDA (eXistDA.eXistDA):
    """An Remote eXist connector object."""

    meta_type = utils.REXISTDA_METATYPE

    #__ac_permissions__ = (
    #  (utils.PERM_MANAGE_ACCESS, ('manage_tabs', 'manage_main', 'manage_edit')),
    #  ('Change permissions', ('manage_access')),
    #)


    manage_options = (
      {'label':'Edit', 'action':'manage_main', 'help':('eXistDA','ReXistDA_add.stx')},
      {'label':'Security', 'action':'manage_access'},
    )

    security = ClassSecurityInfo()
    security.declareObjectProtected('View')

    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'manage_tabs', 'manage_main', 'manage_edit')
    security.declareProtected('Change permissions', 'manage_access')

    security.declarePrivate('password')

    index_html = HTMLFile('dtml/index_RDA', globals())
    manage_main = HTMLFile('dtml/ReXistDAEdit', globals())


    def __init__(self, id, title, server='', port=utils.DEFAULT_PORT, username='', password='', encoding=utils.DEFAULT_ENCODING, xmlrpc_path=utils.DEFAULT_ZOPE_XMLRPC_PATH):
        """Class constructor.

* Returns a new object !!

* 'id' : string containing the id of the Zope object created

* 'title' : string containing the title property of the Zope object

* 'server' : remote zope server name which owns the eXistDA object.

* 'port' : port number used by Zope HTTP server. Default is 8080. Ex : '8080'

* 'username' : login of the user on Zope, with access to target eXistDA. Leave empty for guest (for python2.1). Ex: 'my_user'

* 'password' : password of the selected user. Leave empty if none. Ex : '******' ';-)'

* 'encoding' : encoding of the string which are passed and will be returned by eXistDA object. Default is 'ISO-8859-1', if an empty string is passed. Ex : 'UTF-8'

* 'xmlrpc_path' : address of eXistDA object on remote Zope. Ex : '/my/path/existda'"""

        self.id = str(id)
        self.title = str(title)
        self.server = str(server)
        self.port = str(port)
        self.username = str(username)
        self.password = str(password)
        self.encoding = str(encoding)
        self.xmlrpc_path = str(xmlrpc_path)



    security.declarePrivate('_getConnection')
    def _getConnection(self, xmlrpc_path=None):
        """Opens a connection to the remote Zope server if there's no one available.

* Returns the object which is the proxy to the XMLRPC server"""

        if xmlrpc_path == None:
            xmlrpc_path = self.xmlrpc_path

        if self.username.strip() != '' and self.password.strip() != '':
            server = "http://" + self.username + ':' + self.password +'@' + self.server + ":" + self.port + xmlrpc_path
        else:
            server = "http://" + self.server + ":" + self.port + xmlrpc_path

        ## To do : verify that xmlrpc_path points to an eXistDA object.

        return xmlrpclib.ServerProxy(server, encoding=self.encoding)


    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'findXMLRPCAddress')
    def findXMLRPCAddress(self):
        """Dummy method to provide compatibility to eXistDA. Always return "" for ReXistDA."""

        return ""


    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'isOnline')
    def isOnline(self):
        return (self.id, 1)



    security.declareProtected( utils.PERM_WRITE_ACCESS, 'saveDoc' )
    def saveDoc(self, my_doc, doc_uri, overwrite=1, **kws):
        """Save a document in the database.

* Returns 1 if saved, 0 otherwise

* 'my_doc' : an object/URI/string containing the doc

* 'doc_uri' : a string containing the complete URI (server, path and name) of the doc in the DB

* if 'overwrite' == 1 then if there's an existing doc with the same name, it will be overwritten

* 'kws' : dict for any other parameters. See README.txt for recognized parameters"""

        try:
            return self._getConnection().saveDoc(my_doc, doc_uri, overwrite, kws)
        except:
            return 0

        return 0


    security.declareProtected( utils.PERM_READ_ACCESS, 'getFilesFromQuery')
    def getFilesFromQuery(self, query, **kws):
        """**BROKEN SINCE JUNE 2004 SNAPSHOTS -> modify eXist src to make it work**

Executes an XPATH query or an XQuery on the DB.

* Returns a list of string containing the name and full path of files in the DB which matches the query

* 'query' : a object/URI/string containing an XPATH query or an XQuery

* 'kws' : dict for any other parameters. See README.txt and 'query' for recognized parameters"""

        return self._getConnection().getFilesFromQuery(query, kws)


    security.declareProtected( utils.PERM_READ_ACCESS, 'query' )
    def query(self, query, **kws):
        """Execute an XPATH query or an XQuery on the DB.

* Returns a 'eXistDAresultsSet' object containing all the results returned by the query.

* 'query' : an object/URI/string containing an XPATH query or an XQuery

* 'kws' : dict for any other parameters. See 'README.txt' for recognized parameters

  * 'object_only' : cf. 'README.txt'

  * 'variables' : key to a dict containing variables to pass to the xquery"""

        return self._getConnection().query(query, kws)


    security.declareProtected( utils.PERM_READ_ACCESS, 'isCollection' )
    def isCollection(self, resource_name):
        """Determines if 'resource_name' is a collection in the DB or not.

* Returns 1 if 'resource_name' is a collection, 0 otherwise

* 'resource_name' : a string containing the full path and name of the resource to test"""

        self._getConnection().isCollection(resource_name)


    security.declareProtected( utils.PERM_READ_ACCESS, 'isDocument' )
    def isDocument(self, resource_name):
        """Determines if 'resource_name' is a document in the DB or not.

* Returns 1 if 'resource_name' is a document, 0 otherwise

* 'resource_name' : a string containing the full path and name of the resource to test"""

        return self._getConnection().isDocument(resource_name)


    security.declareProtected( utils.PERM_READ_ACCESS, 'getDoc' )
    def getDoc(self, doc_name, xsl='', pretty_print=utils.DEFAULT_PP, **kws):
        """Retrieve a document in the database.

* Returns a string containing the (XSL transformed if suited) XML document.

* 'doc_name' : a string containing the full path and name of an existing document in the DB

* 'xsl' : a string containing the full path and name of an existing XSL document in the DB. If an empty string is passed, no XSLT is done

* 'if pretty_print' == 1, then the returned string is "pretty printed", ie indented. Else, the string is returned as it is saved in the DB

* 'kws' : dict for any other parameters. See README.txt for recognized parameters"""

        return self._getConnection().getDoc(doc_name, xsl, pretty_print, kws)


    security.declareProtected( utils.PERM_READ_ACCESS, 'getDictFromDoc' )
    def getDictFromDoc(self, doc_name):
        """Returns a dictionary which keys are XML tags and values are XML values from the doc.

**WARNING** : for the moment, attributes are ignored and forgotten.

* Returns a python dictionary

* doc_name : a string containing the full path and name of an existing document in the DB"""

        return self._getConnection().getDictFromDoc(doc_name)


    security.declareProtected( utils.PERM_READ_ACCESS, 'listColl' )
    def listColl(self, coll_name=utils.EXIST_ROOT_COLL, recursive=0, full_path=1):
        """Lists the collections contained in collection coll_name.

* Returns a list containing strings which are sub-collections of 'coll_name'

* 'coll_name' : a string containing the base collection, containing the others. If the parameter is not passed, defaults to 'EXIST_ROOT_COLL' (ie '/db')

* if 'recursive' == 0, returned values will be the direct children collections of 'coll_name'. Otherwise, all sub-collections will be recursively returned (even the children of the children)

* if 'full_path' == 1, the full path to the collections is returned (including '/db/...'). Otherwise, the path returned is relative to 'coll_name'"""

        return self._getConnection().listColl(coll_name, recursive, full_path)


    security.declareProtected( utils.PERM_READ_ACCESS, 'listDoc' )
    def listDoc(self, coll_name='', full_path=0):
        """
        Returns the list of all documents contained in collection 'coll_name'.

- 'coll_name' : the name (with complete path) of the collection you want to list the child collections

- 'full_path' : if != 0 returns the full path of the document, otherwise, only the document name
        """

        return self._getConnection().listDoc(coll_name, full_path)


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'xupdate' )
    def xupdate(self, res_name, xupdate_doc, **kws):
        """Issues an XUpdate command to update a document or a whole collection in the database.

- Returns a number > 0 (nb of updates done ?) if AOK or raise an exception if pb

- 'res_name' : the name (with complete path) of a collection or a document

- 'xupdate_doc' : an object/URI/string containing an XUpdate string

- 'kws' : dict for any other parameters. See README.txt for recognized parameters
        """

        return self._getConnection().xupdate(res_name, xupdate_doc, kws)


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'createColl' )
    def createColl(self, coll_name):
        """Creates a new collection in the coll_name collection.

- Returns 1 if no problem. Returns 0 if collection has not been created

- 'coll_name' : string containing the name of the collection to create, with the **FULL** path
        """

        return self._getConnection().createColl(coll_name)


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'delColl' )
    def delColl(self, coll_name) :
        """Deletes a collection and **ALL** what it contains.

- Returns 1 if no problem. Returns 0 if collection has not been deleted

- 'coll_name' : string containing the name of the collection to delete, with the **FULL** path
        """

        return self._getConnection().delColl(coll_name)


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'delDoc' )
    def delDoc(self, doc_name) :
        """Deletes a document of the DB.

- Returns 1 if no problem. Returns 0 if document has not been deleted

- 'doc_name' : string containing the name of the document to delete, with the **FULL** path
        """

        return self._getConnection().delDoc(doc_name)

InitializeClass(ReXistDA)



"""xslt.py : XSL transformation with local 4Suite or xsltproc in eXistDA"""

# -*- coding: ISO-8859-1 -*-

#    This file is part of eXistDA
#
#    Copyright (C) 2004  Sebastien PILLOZ - Ecole Normale Superieure de Lyon
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


__doc__ = """xslt.py : XSL transformation with local 4Suite or xsltproc in eXistDA"""


SCRIPT = "EXISTDA.xslt.py"

import utils
from zLOG import LOG, INFO, DEBUG

try:
    from Ft.Xml.Xslt import Processor
    from Ft.Xml.InputSource import DefaultFactory
    from Ft.Xml import Domlette, cDomlette
    LOG(SCRIPT + ' :', INFO, "4Suite found. External XSLT will be available for eXistDA.")
except:
    LOG(SCRIPT + ' :', INFO, utils.EXC_NO_4SUITE)

import types


def transform(xslStyleSheet, xmlInput, params=None, xsl_base_uri='', apply_xsl=1, encoding=utils.DEFAULT_ENCODING):
    """Performs an XSL transformation of 'xmlInput' with 'xslStyleSheet'. Returns a string containing the result of the XSLT.

* 'xslStyleSheet' : a string or stream containing the XSL to use

* 'xmlInput' : the XML string to be transformed

* 'params' : a dict containing the parameters to be passed to the xsl

* 'xsl_base_uri' : base uri of the XSL. Must be a string

* 'apply_xsl' : set to 1 (default) to apply the XSLT, 0 to return the XML string not transformed"""

    if utils.USE_XSLTPROC and utils.hasXSLTProc():
        return xsltproc(xslStyleSheet, xmlInput, params, xsl_base_uri, encoding)

    if utils.USE_CDOMLETTE:
        dom_model = cDomlette.implementation
    else:
        dom_model = Domlette.implementation

    if apply_xsl == 1 :
        processor = Processor.Processor(implementation=dom_model, stylesheetIncPaths=[xsl_base_uri])

        if utils.isHTTPURI(xslStyleSheet) or utils.isFileURI(xslStyleSheet):
            ma_xsl = DefaultFactory.fromUri(xslStyleSheet, xsl_base_uri)
        elif type(xslStyleSheet) == types.StringType:
            ma_xsl = DefaultFactory.fromString(xslStyleSheet, xsl_base_uri)
        else:
            ma_xsl = DefaultFactory.fromStream(xslStyleSheet, xsl_base_uri)

        processor.appendStylesheet(ma_xsl)

        mon_xml = DefaultFactory.fromString(xmlInput, xsl_base_uri)

        resultat = processor.run(mon_xml, topLevelParams=params)

        return resultat


    return xmlInput

def xsltproc(xsl, xml, params, uri, encoding=utils.DEFAULT_ENCODING):
    """Do the transformation, using xsltproc"""

    from tempfile import NamedTemporaryFile
    from os import popen
    from time import time

    f = NamedTemporaryFile(mode="wt")
    f.write(xml)

    str_params =""
    if params:
        for key in params:
            str_params += "--stringparam %s %s " % (key, params[key])

    f.flush()
    cmd = "xsltproc --novalid --nonet %s %s %s" % (str_params, xsl.split('://',1)[-1], f.name)
    if (utils.DEBUG > 2):
        LOG(SCRIPT, DEBUG, "xsltproc command line is : \n----------\n%s\n----------" % cmd)
    htmlfile = popen(cmd, "r")

    start_time = time()
    html = htmlfile.read()
    while (len(html) == 0) and (time() - start_time < utils.XSLT_TIMEOUT) :
        html = htmlfile.read()

    if htmlfile.close():
        raise('Unable to complete XSLT with xsltproc')

    if (utils.DEBUG > 2):
        LOG(SCRIPT, DEBUG, "xsltproc returned : \n----------\n%s\n----------" % html)

    f.close()
    # xsltproc produces iso-8859-1 code. Reencode with correct encoding.
    try:
        return unicode(html, 'iso-8859-1').encode(encoding)
    except:
        if (utils.DEBUG > 0):
            LOG(SCRIPT, DEBUG, "xsltproc couldn't return correct encoding")
        return html




"""eXistDA.py : eXist Database Adapter product module for Zope."""

# -*- coding: ISO-8859-1 -*-

#    This file is part of eXistDA.
#
#    Copyright (C) 2004  Sebastien PILLOZ - Ecole Normale Superieure de Lyon
#
#    eXistDA is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


__doc__ = """eXistDA product module."""

from App.special_dtml import HTMLFile
from App.Dialogs import MessageDialog
from Persistence import Persistent
from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass

import OFS.SimpleItem
import Acquisition
import AccessControl.Role
from zLOG import LOG, DEBUG

import xmlrpclib, urllib
from xmlrpclib import Binary
from types import StringType

from utils import iseXistURI, isHTTPURI
import utils
import eXistDAresult
import xslt

import string

## Zope method for management
manage_addeXistDAForm = HTMLFile('dtml/eXistDAAdd', globals())
def manage_addeXistDA(self, id, title='', server='', port='', username='', password='', encoding=utils.DEFAULT_ENCODING, xmlrpc_path=utils.DEFAULT_XMLRPC_PATH, REQUEST=None):
    """Add an eXist connector object."""
    # Avoid empty encoding
    if encoding.strip() == '':
        encoding = utils.DEFAULT_ENCODING

    msg = utils.getInvalidMsg(server, port, username, password, encoding)
    if msg is None :
        self._setObject(id, eXistDA(id, title, server, port, username, password, encoding, xmlrpc_path))
        if REQUEST is not None:
            return self.manage_main(self, REQUEST)
    else:
        if REQUEST is not None:
            return msg


# The class which represents an eXist "proxy" or database adapter.
class eXistDA (OFS.SimpleItem.Item, Persistent, Acquisition.Implicit, AccessControl.Role.RoleManager):
    """An eXist connector object."""

    meta_type = utils.EXISTDA_METATYPE

    manage_options = (
      {'label':'Edit', 'action':'manage_main'},
      {'label':'Security', 'action':'manage_access'},
    )

    security = ClassSecurityInfo()
    security.declareObjectProtected('View')

    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'manage_tabs', 'manage_main', 'manage_edit')
    security.declareProtected('Change permissions', 'manage_access')
    security.declarePrivate('password')
    def __init__(self, id, title, server='', port=utils.DEFAULT_PORT, username='', password='', encoding=utils.DEFAULT_ENCODING, xmlrpc_path=utils.DEFAULT_XMLRPC_PATH):
        """Class constructor.

* Returns a new object !!

* 'id' : string containing the id of the Zope object created

* 'title' : string containing the title property of the Zope object

* 'server' : server name which runs the DB. Should be a complete name and **SHOULD NEVER BE 'LOCALHOST' or '127.0.0.1'** because it may be, sometimes, evaluated from the client within a link. Ex : 'www.mycomputer.org'

* 'port' : port number used by the eXist's XMLRPC server. Default is 8080. Ex : '8080'

* 'username' : login of the user on the DB. Leave empty for guest (for python2.1). Ex: 'my_user'

* 'password' : password of the selected user. Leave empty if none. Ex : '******' ';-)'

* 'encoding' : encoding of the string which are passed and will be returned by the DB. Default is 'ISO-8859-1', if an empty string is passed. Ex : 'UTF-8'

* 'xmlrpc_path' : address of eXist embedded XMLRPC server. Ex : '/xmlrpc'"""

        self.id = str(id)
        self.title = str(title)
        self.server = str(server)
        self.port = str(port)
        self.username = str(username)
        self.password = str(password)
        self.encoding = str(encoding)
        self.xmlrpc_path = str(xmlrpc_path)

    index_html = HTMLFile('dtml/index_DA', globals())
    manage_main = HTMLFile('dtml/eXistDAEdit', globals())

    security.declareProtected(utils.PERM_READ_ACCESS, 'getDA')
    def getDA(self):
        """Returns a reference to the object itself"""
        return self.__of__

    ###############################################################################
    ### XMLRPC & eXist methods

    ## Private methods, users dont have to see them
    security.declarePrivate('_getConnection')
    def _getConnection(self, xmlrpc_path=None):
        """Opens a connection to the eXist server if there's no one available.

* Returns the object which is the proxy to the XMLRPC server"""

        # See why this bugs...
        #if self.connected is None:
        #    server = "http://" + self.server + ":" + self.port + "/exist/xmlrpc"
        #    self.connected = xmlrpclib.ServerProxy(server, encoding=self.encoding)
        if xmlrpc_path == None:
            xmlrpc_path = self.xmlrpc_path

        if self.username.strip() != '' and self.password.strip() != '':
            server = "http://" + self.username + ':' + self.password +'@' + self.server + ":" + self.port + xmlrpc_path
        else:
            server = "http://" + self.server + ":" + self.port + xmlrpc_path

        return xmlrpclib.ServerProxy(server, encoding=self.encoding)

    security.declarePrivate('_serverIsMe')
    def _serverIsMe(self, URI, **kws):
        """Returns 1 if 'URI' "represents" this object.

* 'URI' : a string containing an URI

* 'kws' : dict for any other parameters. See README.txt for recognized parameters"""

        server = utils.getServerNameFromURI(URI)
        if server.find(':') > -1:
            server = server.split(':')[server.count(':') -1].strip()
        if server.find('@') > -1:
            server = server.split('@')[1].strip()
        # if no server name is passed, assume I'm the good one
        if server.strip() == '':
            return 1

        if server.find('.') > -1:
            if server == self.server:
                return 1
        else:
            if server == self.server.split('.')[0]:
                return 1

        return 0


    security.declarePrivate('_getAnythingAsString')
    def _getAnythingAsString(self, my_data, object_only=0):
        """Returns 'my_data' as a string. Raise an error if nothing can be returned. 1 level depth only (ie an HTTP URI in a file will return the HTTP URI, not the string contained in the page at the URI)

* 'my_data' : a string, a file like object, a zope object, an HTTP URI, an eXist URI

* 'object_only' : an integer, which determines if we only do a test for an object (1) or not (0). Useful for URI-like parameters

"""

        try:
            # is my_data a file like object ? Does it have a read method ?
            if hasattr(my_data, 'read'):
                return my_data.read()
            elif hasattr(my_data, 'view_image_or_file'):
                return str(my_data)
        except:
            pass

        if not object_only :
            #try:
                # is my_data a string ?
                if type(my_data) == StringType:
                    # is my_data an HTTP URI ?

                    if isHTTPURI(my_data):
                        return urllib.urlopen(my_data).read()

                    elif iseXistURI(my_data):
                        if self._serverIsMe(my_data, object_only=1):
                            return self.getDoc(my_data, object_only=1)

                        else:
                            return my_data

                    # else, we return the string !
                    else :
                        return my_data

            #except:
                #pass

        else:
            return str(my_data)

        # and the rest ?

        return utils.EXC_NOT_STRING



    ## User's methods (or manager's...)

    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'isOnline')
    def isOnline(self):
        """Determines (by trying a sync() on the DB if the server is online.

* Returns a tuple consisting of the id of the server, and 1 if it's up, 0 if not"""

        try:
            # We do a synchronisation to see if the server is up
            self._getConnection().sync()
            return (self.id, 1)
        except:
            pass

        return (self.id, 0)


    security.declareProtected(utils.PERM_MANAGE_ACCESS, 'findXMLRPCAddress')
    def findXMLRPCAddress(self):
        """Tries to find the address of the eXist XMLRPC embedded server.

* Returns the first address if found, or an empty string if nothing is found."""


        lst_addr = ['/exist/xmlrpc', '/xmlrpc']


        valid_addr = ""
        for addr in lst_addr:
            try:
                self._getConnection(addr).sync()
                valid_addr = addr
                break
            except:
                pass

        return valid_addr


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'saveDoc' )
    def saveDoc(self, my_doc, doc_uri, overwrite=1, **kws):
        """Save a document in the database.

* Returns 1 if saved, 0 otherwise

* 'my_doc' : an object/URI/string containing the doc

* 'doc_uri' : a string containing the complete URI (server, path and name) of the doc in the DB

* if 'overwrite' == 1 then if there's an existing doc with the same name, it will be overwritten

* 'kws' : dict for any other parameters. See README.txt for recognized parameters"""

        err = 0

        if kws.has_key(utils.OBJ_ONLY):
            doc_string = self._getAnythingAsString(my_doc, int(kws[utils.OBJ_ONLY]))
        else:
            doc_string = self._getAnythingAsString(my_doc)


        if self._serverIsMe(doc_uri, object_only=1):
            try:
                # To do : overwrite = 0 seems not to work
                err = self._getConnection().parse(Binary(doc_string), utils.getRscNameFromURI(doc_uri), overwrite)
            except:
                import traceback
                traceback.print_exc()
                raise utils.EXC_CANT_SAVE

        else:
            raise utils.EXC_NOT_THIS_DB

        # err is an XMLRPC Boolean... convert it to a 1 or 0
        return err == 1

    security.declareProtected( utils.PERM_WRITE_ACCESS, 'saveDoc' )
    def storeBinary(self, my_doc, doc_uri, overwrite=1, mimetype="application/binary", **kws):
        """Save a binary document in the database.

* Returns 1 if saved, 0 otherwise

* 'my_doc' : an object/URI/string containing the doc

* 'doc_uri' : a string containing the complete URI (server, path and name) of the doc in the DB

* if 'overwrite' == 1 then if there's an existing doc with the same name, it will be overwritten

* 'kws' : dict for any other parameters. See README.txt for recognized parameters"""

        err = 0

        if kws.has_key(utils.OBJ_ONLY):
            doc_string = self._getAnythingAsString(my_doc, int(kws[utils.OBJ_ONLY]))
        else:
            doc_string = self._getAnythingAsString(my_doc)

        if self._serverIsMe(doc_uri, object_only=1):
            try:
                err = self._getConnection().storeBinary(Binary(doc_string), utils.getRscNameFromURI(doc_uri), mimetype,1==1)
            except:
                import traceback
                traceback.print_exc()
                raise utils.EXC_CANT_SAVE

        else:
            raise utils.EXC_NOT_THIS_DB

        # err is an XMLRPC Boolean... convert it to a 1 or 0
        return err == 1



    security.declareProtected( utils.PERM_READ_ACCESS, 'getFilesFromQuery')
    def getFilesFromQuery(self, query, **kws):
        """**BROKEN SINCE JUNE SNAPSHOTS -> modify eXist src to make it work**

Executes an XPATH query or an XQuery on the DB.

* Returns a list of string containing the name and full path of files in the DB which matches the query

* 'query' : a object/URI/string containing an XPATH query or an XQuery

* 'kws' : dict for any other parameters. See README.txt and 'query' for recognized parameters"""

        xmlrpc_server = self._getConnection()

        if kws.has_key(utils.OBJ_ONLY):
            query = self._getAnythingAsString(query, int(kws[utils.OBJ_ONLY]))
        else:
            query = self._getAnythingAsString(query)

        ns = {}
        if kws.has_key(utils.OK_NAMESPACES):
            ns = kws[utils.OK_NAMESPACES]


        if kws.has_key(utils.KEY_VARIABLES):
            result = xmlrpc_server.executeQuery(Binary(query), self.encoding, {utils.KEY_VARIABLES:kws[utils.KEY_VARIABLES], utils.OK_NAMESPACES:ns})
        else:
            res = xmlrpc_server.executeQuery(Binary(query), self.encoding, {utils.OK_NAMESPACES:ns})

        dico_res = xmlrpc_server.querySummary(res)
        xmlrpc_server.releaseQueryResult(res)
        lst_ret = []

        for i in range(0,len(dico_res['documents'])):
            lst_ret.append(dico_res['documents'][i][0])

        return lst_ret


    security.declareProtected( utils.PERM_READ_ACCESS, 'query' )
    def query(self, query, **kws):
        """Execute an XPATH query or an XQuery on the DB.

* Returns a 'eXistDAresultsSet' object containing all the results returned by the query.

* 'query' : an object/URI/string containing an XPATH query or an XQuery

* 'kws' : dict for any other parameters. See 'README.txt' for recognized parameters

  * 'object_only' : cf. 'README.txt'

  * 'variables' : key to a dict containing variables to pass to the xquery"""

        xmlrpc_server = self._getConnection()

        if kws.has_key(utils.OBJ_ONLY) and kws[utils.OBJ_ONLY] == 1 :
            query = self._getAnythingAsString(query, int(kws[utils.OBJ_ONLY]))
        else:
            query = self._getAnythingAsString(query)


        ns = {}
        if kws.has_key(utils.OK_NAMESPACES):
            ns = kws[utils.OK_NAMESPACES]


        if utils.DEBUG >= 1:
            LOG("eXistDA - query :", DEBUG, "Query is :\n%s" % query)

        # do we need to pass some variables to the xquery ?
        if kws.has_key(utils.KEY_VARIABLES):
            result = xmlrpc_server.executeQuery(Binary(query), self.encoding, {utils.KEY_VARIABLES:kws[utils.KEY_VARIABLES], utils.OK_NAMESPACES:ns})
        else:
            result = xmlrpc_server.executeQuery(Binary(query), self.encoding, {utils.OK_NAMESPACES:ns})

        resSet = eXistDAresult.eXistDAresultsSet(encoding=self.encoding)

        if int(xmlrpc_server.getHits(result)) > 0:
            for i in range(int(xmlrpc_server.getHits(result))):
                if kws.has_key(utils.KEY_VARIABLES):
                    resSet.append(eXistDAresult.eXistDAresult(self.getDA(), xmlrpc_server.retrieve(result, i, {utils.OK_ENCODING:self.encoding, utils.KEY_VARIABLES:kws[utils.KEY_VARIABLES], utils.OK_NAMESPACES:ns}).data, '', 1, self.encoding))
                else:
                    resSet.append(eXistDAresult.eXistDAresult(self.getDA(), xmlrpc_server.retrieve(result, i, {utils.OK_ENCODING:self.encoding, utils.OK_NAMESPACES:ns}).data, '', 1, self.encoding))

        if utils.DEBUG >= 1:
            LOG("eXistDA - query :", DEBUG, "Got a resultsSet for %s" % self.server)
            if utils.DEBUG >= 3:
                LOG("eXistDA - query :", DEBUG, "XML result is :\n%s" % str(resSet.getMergedXML()))

        xmlrpc_server.releaseQueryResult(result)
        if utils.DEBUG >= 3:
                LOG("eXistDA - query :", DEBUG, "Returning resultsSet to caller")
        return resSet



    security.declareProtected( utils.PERM_READ_ACCESS, 'isCollection' )
    def isCollection(self, resource_name):
        """Determines if 'resource_name' is a collection in the DB or not.

* Returns 1 if 'resource_name' is a collection, 0 otherwise

* 'resource_name' : a string containing the full path and name of the resource to test"""

        if resource_name.strip() == '':
            return 0

        # the collection name can't have a / at the end
        if resource_name.strip()[-1:] == '/':
            resource_name = resource_name.strip()[:-1]

        if iseXistURI(resource_name):
            resource_name = utils.getRscNameFromURI(resource_name)

        if resource_name.strip()[0] != '/':
            resource_name = "/" + resource_name.strip()

        if resource_name == utils.EXIST_ROOT_COLL:
            return 1

        parent_coll = '/'.join(resource_name.split('/')[:-1]) + '/'
        this_coll = ''.join(resource_name.split('/')[-1:])

        try:
            if this_coll in self._getConnection().getCollectionDesc(parent_coll)['collections']:
                return 1
        except:
            return 0

        return 0


    security.declareProtected( utils.PERM_READ_ACCESS, 'isDocument' )
    def isDocument(self, resource_name):
        """Determines if 'resource_name' is a document in the DB or not.

* Returns 1 if 'resource_name' is a document, 0 otherwise

* 'resource_name' : a string containing the full path and name of the resource to test"""

        try:
            if iseXistURI(resource_name):
                resource_name = utils.getRscNameFromURI(resource_name)

            if self._getConnection().hasDocument(resource_name):
                return 1
        except:
            pass

        return 0

    security.declareProtected( utils.PERM_WRITE_ACCESS, 'moveCol' )
    def moveCol(self, collectionPath, destinationPath, **kws):
        """ Move a collection """
        print "ici"
        if iseXistURI(collectionPath):
            collectionPath_shortname = utils.getRscNameFromURI(collectionPath)
            print "collectionPath",collectionPath
            newdestinationPath_shortname  = utils.getRscNameFromURI('/'.join(destinationPath.split('/')[:-1]))
            print "newdestinationPath",newdestinationPath_shortname
            newName = destinationPath.split('/')[-1]
            print "newName",newName
            res = self._getConnection().moveCollection(collectionPath_shortname, newdestinationPath_shortname, newName)
            print "res",res
            return res
        else:
            return None

    security.declareProtected( utils.PERM_READ_ACCESS, 'getBin' )
    def getBin(self, doc_name, **kws):
        """ REtrieve a binary ressource """
        if iseXistURI(doc_name):
            doc_name_short = utils.getRscNameFromURI(doc_name)
            bin = self._getConnection().getBinaryResource(doc_name_short).data
            return bin
        else:
            return None

    security.declareProtected( utils.PERM_READ_ACCESS, 'getDoc' )
    def getDoc(self, doc_name, xsl='', pretty_print=utils.DEFAULT_PP, **kws):
        """Retrieve a document in the database.

* Returns a string containing the (XSL transformed if suited) XML document.

* 'doc_name' : a string containing the full path and name of an existing document in the DB

* 'xsl' : a string containing the full path and name of an existing XSL document in the DB. If an empty string is passed, no XSLT is done

* 'if pretty_print' == 1, then the returned string is "pretty printed", ie indented. Else, the string is returned as it is saved in the DB

* 'kws' : dict for any other parameters. See README.txt for recognized parameters"""

        if iseXistURI(doc_name):
            doc_name_short = utils.getRscNameFromURI(doc_name)
            params = {}

            if pretty_print == 1:
                params[utils.OK_INDENT] = 'yes'
                params[utils.OK_INDENT_SPACES] = 'yes'
            else:
                params[utils.OK_INDENT] = 'no'
                params[utils.OK_INDENT_SPACES] = 'no'

            params[utils.OK_ENCODING] = self.encoding


            if iseXistURI(xsl):
                if kws.has_key(utils.DICT_STYLESHEET_PARAM):
                    params[utils.OK_STYLESHEET_PARAM] = kws[utils.DICT_STYLESHEET_PARAM]

                xsl_short = '/' + xsl.strip().split('//')[1].split('/',1)[1]
                params[utils.OK_STYLESHEET] = xsl_short

                return self._getConnection().getDocument(doc_name_short, params).data
            elif type(xsl) == StringType and xsl.strip() == '':
                return self._getConnection().getDocument(doc_name_short, params).data
            else:
                ma_xsl = self._getAnythingAsString(xsl)
                mon_xml = self._getAnythingAsString(doc_name)
                if utils.DEBUG > 2:
                    LOG('eXistDA - getDoc', DEBUG, "xsl is : %s\n\n xml is %s\n" % (ma_xsl, mon_xml))
                return xslt.transform(ma_xsl, mon_xml, encoding=self.encoding)

        # the doc isn't on an eXist DB
        else:
            params = None
            xsl_base_uri = ''
            if kws.has_key(utils.DICT_STYLESHEET_PARAM):
                params = kws[utils.DICT_STYLESHEET_PARAM]
            if kws.has_key(utils.DICT_STYLESHEET_BASE_URI):
                xsl_base_uri = kws[utils.DICT_STYLESHEET_BASE_URI]

            ma_xsl = xsl
            if kws.has_key(utils.OBJ_ONLY):
                mon_xml = self._getAnythingAsString(doc_name, object_only=kws[utils.OBJ_ONLY])
            else:
                mon_xml = self._getAnythingAsString(doc_name)

            return xslt.transform(ma_xsl, mon_xml, params, xsl_base_uri=xsl_base_uri, encoding=self.encoding)


        #else:
        #    raise utils.EXC_DOC_MUST_BE_ON_EXIST_SERVER



    security.declareProtected( utils.PERM_READ_ACCESS, 'getDictFromDoc' )
    def getDictFromDoc(self, doc_name):
        """Returns a dictionary which keys are XML tags and values are XML values from the doc.

**WARNING** : for the moment, attributes are ignored and forgotten.

* Returns a python dictionary

* doc_name : a string containing the full path and name of an existing document in the DB"""

        if iseXistURI(doc_name) :
            doc = self.getDoc(doc_name, xsl='')
        else:
            doc = self._getAnythingAsString(doc_name)


        if doc.strip() != '':
            dict = utils.getDict(doc, self.encoding)
            return dict

        return {}



    security.declareProtected( utils.PERM_READ_ACCESS, 'listColl' )
    def listColl(self, coll_name=utils.EXIST_ROOT_COLL, recursive=0, full_path=1):
        """Lists the collections contained in collection coll_name.

* Returns a list containing strings which are sub-collections of 'coll_name'

* 'coll_name' : a string containing the base collection, containing the others. If the parameter is not passed, defaults to 'EXIST_ROOT_COLL' (ie '/db')

* if 'recursive' == 0, returned values will be the direct children collections of 'coll_name'. Otherwise, all sub-collections will be recursively returned (even the children of the children)

* if 'full_path' == 1, the full path to the collections is returned (including '/db/...'). Otherwise, the path returned is relative to 'coll_name'"""


        liste_def=[]

        if iseXistURI(coll_name):
            coll_name = utils.getRscNameFromURI(coll_name)

        if coll_name.strip()[-1:] != '/':
            coll_name = coll_name.strip() + "/"

        if recursive == 0:
            liste_coll = self._getConnection().getCollectionDesc(coll_name)['collections']

            if full_path == 1:
                for une_coll in liste_coll:
                    liste_def.append(coll_name + une_coll + '/')
            else:
                return liste_coll

        else:
            liste_def.append(coll_name)
            desc = self._getConnection().getCollectionDesc(coll_name)

            if desc != None:
                for collection in desc['collections']:
                    liste_def = liste_def + self.listColl(coll_name + collection + '/', recursive, full_path)

        return liste_def

    security.declareProtected( utils.PERM_READ_ACCESS, 'listDoc' )
    def listDoc(self, coll_name='', full_path=0):
        """
        Returns the list of all documents contained in collection 'coll_name'.

- 'coll_name' : the name (with complete path) of the collection you want to list the child collections

- 'full_path' : if != 0 returns the full path of the document, otherwise, only the document name
        """

        if iseXistURI(coll_name):
            coll_name = utils.getRscNameFromURI(coll_name)

        xmlrpc_server = self._getConnection()

        if coll_name.strip() != '':
            if coll_name[-1:] != '/':
                coll_name = coll_name + '/'

            liste_tmp = xmlrpc_server.getDocumentListing(coll_name)
        else:
            liste_tmp =  xmlrpc_server.getDocumentListing()

        if coll_name == '' :
            coll_name = utils.EXIST_ROOT_COLL

        if full_path == 0 :
            liste = liste_tmp
        else:
            liste = []
            for item in liste_tmp:
                liste.append(coll_name + item)

        return liste


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'xupdate' )
    def xupdate(self, res_name, xupdate_doc, **kws):
        """Issues an XUpdate command to update a document or a whole collection in the database.

- Returns a number > 0 (nb of updates done ?) if AOK or raise an exception if pb

- 'res_name' : the name (with complete path) of a collection or a document

- 'xupdate_doc' : an object/URI/string containing an XUpdate string

- 'kws' : dict for any other parameters. See README.txt for recognized parameters
        """

        xmlrpc_server = self._getConnection()

        if kws.has_key(utils.OBJ_ONLY):
            xupdate_doc = self._getAnythingAsString(xupdate_doc, int(kws[utils.OBJ_ONLY]))
        else:
            xupdate_doc = self._getAnythingAsString(xupdate_doc)

        if iseXistURI(res_name):
            res_name = utils.getRscNameFromURI(res_name)

        if self.isCollection(res_name):
            # workaround cause there's no encoding choice for xupdate RPC API function...
            xupdate_doc = utils.convert(xupdate_doc, self.encoding, 'UTF-8')

            return xmlrpc_server.xupdate(res_name, Binary(xupdate_doc))
        elif self.isDocument(res_name):
            return xmlrpc_server.xupdateResource(res_name, Binary(xupdate_doc),self.encoding)
        else:
            raise utils.EXC_NOT_RESOURCE

        return 0


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'createColl' )
    def createColl(self, coll_name):
        """Creates a new collection in the coll_name collection.

- Returns 1 if no problem. Returns 0 if collection has not been created

- 'coll_name' : string containing the name of the collection to create, with the **FULL** path
        """

        if iseXistURI(coll_name):
            coll_name = utils.getRscNameFromURI(coll_name)

        try:
            if self._getConnection().createCollection(coll_name):
                return 1
        except:
            return 0

        return 0


    security.declareProtected( utils.PERM_WRITE_ACCESS, 'delColl' )
    def delColl(self, coll_name) :
        """Deletes a collection and **ALL** what it contains.

- Returns 1 if no problem. Returns 0 if collection has not been deleted

- 'coll_name' : string containing the name of the collection to delete, with the **FULL** path
        """

        if coll_name[-1:] == '/':
            coll_name = coll_name[:-1]

        try:
            if self._getConnection().removeCollection(coll_name):
                return 1
        except:
            return 0

        return 0

    security.declareProtected( utils.PERM_WRITE_ACCESS, 'delDoc' )
    def delDoc(self, doc_name) :
        """Deletes a document of the DB.

- Returns 1 if no problem. Returns 0 if document has not been deleted

- 'doc_name' : string containing the name of the document to delete, with the **FULL** path
        """

        if iseXistURI(doc_name):
            doc_name = utils.getRscNameFromURI(doc_name)

        try:
            if self._getConnection().remove(doc_name):
                return 1
        except:
            return 0

        return 0

    security.declareProtected(  utils.PERM_WRITE_ACCESS, 'delRsc' )
    def delRsc(self, rsc_name) :
        """Wrapper to delColl or delDoc. Automatically chooses the good one (TM)

- Returns 1 if no problem. Returns 0 if resource has not been deleted

- 'rsc_name' : string containing the name of a document or collection that must be deleted, with the **FULL** path"""

        if iseXistURI(rsc_name):
            rsc_name = utils.getRscNameFromURI(rsc_name)

        if self.isCollection(rsc_name):
            return self.delColl(rsc_name)
        elif self.isDocument(rsc_name):
            return self.delDoc(rsc_name)

        return 0



    ## End XMLRPC & eXist methods
    ##########################################################################################

    security.declareProtected( utils.PERM_MANAGE_ACCESS, 'manage_edit' )
    def manage_edit(self, title="", server="", port=utils.DEFAULT_PORT, username="", password="", encoding=utils.DEFAULT_ENCODING, xmlrpc_path=utils.DEFAULT_XMLRPC_PATH, REQUEST=None):
        """Change values of object's parameters. (see '__init__')"""
        self.title = str(title)

        # Avoid empty encoding
        if encoding.strip() == '':
            encoding = utils.DEFAULT_ENCODING

        msg = utils.getInvalidMsg(server, port, username, password, encoding)

        if msg is None:

            self.server = str(server).strip()
            self.port = str(port).strip()
            self.username = str(username).strip()
            self.password = str(password).strip()
            self.encoding = str(encoding).strip()
            self.xmlrpc_path = str(xmlrpc_path).strip()



            if REQUEST is not None:
                return MessageDialog(
                    title = 'Edited',
                    message = "Properties for %s changed." % self.id,
                    action = './manage_main',
                    )

        else:
            if REQUEST is not None:
                return msg


    security.declareProtected( utils.PERM_READ_ACCESS, 'fastquery' )
    def fastquery(self, query, start, nbmax, **kws):
        """Execute an XPATH query or an XQuery on the DB.

        * Returns a 'eXistDAresultsSet' object containing all the results returned by the query.

        * 'query' : an object/URI/string containing an XPATH query or an XQuery

        * 'kws' : dict for any other parameters. See 'README.txt' for recognized parameters

        * 'object_only' : cf. 'README.txt'

        * 'variables' : key to a dict containing variables to pass to the xquery"""

        xmlrpc_server = self._getConnection()

        if kws.has_key(utils.OBJ_ONLY) and kws[utils.OBJ_ONLY] == 1 :
            query = self._getAnythingAsString(query, int(kws[utils.OBJ_ONLY]))
        else:
            query = self._getAnythingAsString(query)

        ns = {}
        if kws.has_key(utils.OK_NAMESPACES):
            ns = kws[utils.OK_NAMESPACES]

        if utils.DEBUG >= 1:
            LOG("eXistDA - query :", DEBUG, "Query is :\n%s" % query)

        # do we need to pass some variables to the xquery ?
        if kws.has_key(utils.KEY_VARIABLES):
            result = xmlrpc_server.query(Binary(query), nbmax, start, {utils.KEY_VARIABLES:kws[utils.KEY_VARIABLES], utils.OK_NAMESPACES:ns})
        else:
            result = xmlrpc_server.query(Binary(query), nbmax, start, {utils.OK_NAMESPACES:ns})


        # replace invalid header result if not result
        res = string.replace(result.data, '<?xml version="1.0"?>', '')
        if utils.DEBUG >= 1:
            LOG("eXistDA - query :", DEBUG, "Got a resultsSet for %s" % self.server)
            if utils.DEBUG >= 3:
                LOG("eXistDA - query :", DEBUG, "XML result is :\n%s" % str(resSet.getMergedXML()))


        if utils.DEBUG >= 3:
                LOG("eXistDA - query :", DEBUG, "Returning resultsSet to caller")
        # return raw xml data set
        return res

InitializeClass(eXistDA)
